float planar_clouds_density(clouds_2D_settings pc, vec3 position){
    const float offset = 1.0 / pc.scale_mask;
    const float scale = 1.0 / (8.0 * pc.scale_mask);

    float altitude = (position.y - pc.altitude_min) / pc.thickness;
    float animation = frameTimeCounter * pc.time_shift;

    vec3 direction0 = normalize(vec3(-1.0, 0.0, 1.0));
    vec3 direction1 = normalize(vec3(-1.0, 0.5, 1.0));

    float n0 = perlin_fbm(vec3(position.x + offset, position.y, position.z) * scale, pc.octaves0);
    float n1 = perlin_fbm(vec3(position.x, position.y + offset, position.z) * scale, pc.octaves0);
    float n2 = perlin_fbm(vec3(position.x, position.y, position.z + offset) * scale, pc.octaves0);

    position = vec3(position.x * n0, position.y * n1, position.z * n2);
    
    vec3 position0 = (position / pc.scale_mask) + animation * direction0; position0.x *= 0.4;
    vec3 position1 = (position + animation * direction1) / pc.scale_detail;

    float noise0 = perlin_fbm(position0, pc.octaves0);
          noise0 = noise0 * pc.coverage * 2.0 - 1.2;

    float noise1 = perlin_fbm(position1, pc.octaves1) * 0.5;
	
	if(noise0 <= 0.0) return 0.0;
	if(noise1 <= 0.0) return 0.0;

    float result = 2.0 * saturate(noise0 - noise1);
          result = pow3(result);

    return result * pc.density;
}  

float planar_clouds_transmittance(clouds_2D_settings pc, vec3 position, vec3 direction, const int steps){
    float ray_length = 64.0;
    vec3 ray_increment = direction * ray_length;
    vec3 ray_position = position + ray_increment * 0.5;
            
    float transmittance = 0.0;

    for(int i = 0; i < steps; i++){
        float density = ray_length * planar_clouds_density(pc, ray_position);
        if(density <= 0.0) continue;
                  
        ray_position += ray_increment * (i + 0.5);
        transmittance += ray_position.y > pc.altitude_max ? 0.0 : density;
        ray_length *= 1.5;
    }

    return transmittance;
}

vec4 planar_clouds(vec3 position, vec3 direction, vec3 sun_direction, vec3 moon_direction, vec3 atmosphere){
    atmosphere_constant ac = atmosphere_s();
    clouds_2D_settings pc = s_clouds2D();

    vec3 clouds_scattering;
    float transmittance = 1.0;

    vec3 up_position = vec3(0.0, 1.0, 0.0);
    vec3 view_position = vec3(0.0, ac.planet_radius + cameraPosition.y, 0.0);
    
    float VdotL = dot(position, direction);
    float VdotU = dot(position, up_position);

    vec2 sphere = ray_sphere_intersection(view_position, position, ac.planet_radius);
    vec2 sphere_min = ray_sphere_intersection(view_position, position, ac.planet_radius + pc.altitude_min);
    vec2 sphere_max = ray_sphere_intersection(view_position, position, ac.planet_radius + pc.altitude_max);

    if(transmittance <= pc.treshold) return vec4(0.0, 0.0, 0.0, 1.0);
    if((eyeAltitude < pc.altitude_min && sphere.y > 0.0) || (eyeAltitude > pc.altitude_max && position.y > 0)) return vec4(0.0, 0.0, 0.0, 1.0);

    vec3 clouds_position = position * sphere_min.y + cameraPosition;
		 clouds_position = clouds_curvature(clouds_position);

    float density = planar_clouds_density(pc, clouds_position) * pc.thickness;
    if(density <= 0.0) return vec4(0.0, 0.0, 0.0, 1.0);

    float step_transmittance = exp(-density);
    float step_transmittance_fraction = 1.0 - step_transmittance;

    float direct_transmittance = planar_clouds_transmittance(pc, clouds_position, direction, 6);
    float indirect_transmittance = planar_clouds_transmittance(pc, clouds_position, up_position, 4) * 0.1; 

    float powder = 1.0 - 0.97 * exp(-4.0 * density);
	float a = 1.0, b = 1.0, c = 1.0;
	
    for(int i = 0; i < 8; ++i){
        float clouds_transmittance = transmittance * (step_transmittance_fraction * a);

        float step_direct_transmittance = clouds_transmittance * exp(-direct_transmittance * b);
        float step_indirect_transmittance = clouds_transmittance * exp(-indirect_transmittance * b);

        float phase0 = volumetric_clouds_phase(VdotL, c);

        clouds_scattering.x += step_direct_transmittance * powder * phase0;
        clouds_scattering.y += step_indirect_transmittance * powder;
		
		a *= 0.6;
		b *= 0.3;
		c *= 0.8;
    }

    transmittance *= step_transmittance;
    transmittance = saturate((transmittance - pc.treshold) / (1.0 - pc.treshold));

    vec3 sunlight = atmospheric_transmittance(sun_direction, moon_direction);
    vec3 skylight = texture2(colortex1, project_sphere(up_position)).xyz;

    vec3 transmittance0 = atmospheric_transmittance(ac, vec3(0.0, ac.planet_radius + cameraPosition.y, 0.0), position, direction);
    vec3 transmittance1 = atmospheric_transmittance(ac, vec3(0.0, ac.planet_radius + cameraPosition.y, 0.0) + position * sphere_min.y, position, direction);
	
    vec3 scattering = clouds_scattering.x * sunlight;
         scattering += clouds_scattering.y * skylight;
		 scattering = mix(atmosphere * (1.0 - transmittance), scattering, saturate(transmittance0 / transmittance1));

    #if ATMOSPHERE_TYPE == 0
        return vec4(scattering, transmittance);
    #elif ATMOSPHERE_TYPE == 1
        return vec4(0.0, 0.0, 0.0, 1.0);
    #endif
}